# -*- coding: UTF-8 -*-
# Time: 10/07/18
# File: tfgantrainer.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals


import tensorflow as tf
import tensorflow.contrib.slim as slim
from antgo.trainer.trainer import *
from antgo.utils import logger
from antgo.measures.moving_statistic import *
from antgo.context import *
from antgo.trainer import tfmodel_deploy
import functools
import os
import re


class TFGANTrainer(Trainer):
  def __init__(self, dump_dir, is_training=True):
    super(TFGANTrainer, self).__init__(is_training)
    self.dump_dir = dump_dir
    self.time_stat = MovingAverage(self.log_every_n_steps)

    self.coord = None
    self.threads = None

    self.saver = None
    self.sess = None
    self.graph = None

    self.loss_list = {}
    self.update_list = {}
    self.lr_list = {}
    self.optimizer_list = {}
    self.var_list = {}
    self.global_step_list = {}

    self.trainop_list = {}
    self._has_model_input = False

    self.cache = {}
    self.loss_log = {}

    self.summary_op = None
    self.train_writer = None

    self.log_iter_at = {}

  def configure_optimizer(self, learning_rate):
    """Configures the optimizer used for training.

    Args:
      learning_rate: A scalar or `Tensor` learning rate.

    Returns:
      An instance of an optimizer.

    Raises:
      ValueError: if optimizer is not recognized.
    """
    if self.optimizer == 'adadelta':
      optimizer = tf.train.AdadeltaOptimizer(
        learning_rate,
        rho=self.adadelta_rho,
        epsilon=self.opt_epsilon)
    elif self.optimizer == 'adagrad':
      optimizer = tf.train.AdagradOptimizer(
        learning_rate,
        initial_accumulator_value=self.adagrad_initial_accumulator_value)
    elif self.optimizer == 'adam':
      optimizer = tf.train.AdamOptimizer(
        learning_rate,
        beta1=self.adam_beta1,
        beta2=self.adam_beta2,
        epsilon=self.opt_epsilon)
    elif self.optimizer == 'ftrl':
      optimizer = tf.train.FtrlOptimizer(
        learning_rate,
        learning_rate_power=self.ftrl_learning_rate_power,
        initial_accumulator_value=self.ftrl_initial_accumulator_value,
        l1_regularization_strength=self.ftrl_l1,
        l2_regularization_strength=self.ftrl_l2)
    elif self.optimizer == 'momentum':
      optimizer = tf.train.MomentumOptimizer(
        learning_rate,
        momentum=self.momentum,
        name='Momentum')
    elif self.optimizer == 'rmsprop':
      optimizer = tf.train.RMSPropOptimizer(
        learning_rate,
        decay=self.rmsprop_decay,
        momentum=self.rmsprop_momentum,
        epsilon=self.opt_epsilon)
    elif self.optimizer == 'sgd':
      optimizer = tf.train.GradientDescentOptimizer(learning_rate)
    else:
      raise ValueError('Optimizer [%s] was not recognized', self.optimizer)
    return optimizer

  def get_variables_to_train(self):
    """Returns a list of variables to train.

    Returns:
      A list of variables to train by the optimizer.
    """
    trainable_scopes = getattr(self, 'trainable_scopes', None)
    if trainable_scopes is None:
      trainable_filter = getattr(self, 'trainable_filter', None)
      if trainable_filter is None:
        return tf.trainable_variables()
      else:
        return [var for var in tf.trainable_variables() if var.name.startswith(trainable_filter)]
    else:
      scopes = [scope.strip() for scope in trainable_scopes.split(',')]

    variables_to_train = []
    for scope in scopes:
      variables = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope)
      variables_to_train.extend(variables)
    return variables_to_train

  def configure_learning_rate(self, num_samples_per_epoch, global_step):
    """Configures the learning rate.

    Args:
      num_samples_per_epoch: The number of samples in each epoch of training.
      global_step: The global_step tensor.

    Returns:
      A `Tensor` representing the learning rate.

    Raises:
      ValueError: if
    """
    batch_size = getattr(self, 'batch_size', 1) * getattr(self, 'num_clones', 1)
    num_epochs_per_decay = getattr(self, 'num_epochs_per_decay', 10)
    decay_steps = int(num_samples_per_epoch / batch_size * num_epochs_per_decay)

    sync_replicas = getattr(self, 'sync_replicas', False)
    if sync_replicas:
      replicas_to_aggregate = getattr(self, 'replicas_to_aggregate', 1)
      decay_steps /= replicas_to_aggregate

    learning_rate_decay_type = getattr(self, 'learning_rate_decay_type', 'exponential')
    learning_rate = getattr(self, 'learning_rate', 0.1)
    if learning_rate_decay_type == 'exponential':
      learning_rate_decay_factor = getattr(self, 'learning_rate_decay_factor', 0.9)
      return tf.train.exponential_decay(learning_rate,
                                        global_step,
                                        decay_steps,
                                        learning_rate_decay_factor,
                                        staircase=True,
                                        name='exponential_decay_learning_rate')
    elif learning_rate_decay_type == 'fixed':
      return tf.constant(learning_rate, name='fixed_learning_rate')
    elif learning_rate_decay_type == 'polynomial':
      end_learning_rate = getattr(self, 'end_learning_rate', 0.0001)
      return tf.train.polynomial_decay(learning_rate,
                                       global_step,
                                       decay_steps,
                                       end_learning_rate,
                                       power=0.9,
                                       cycle=False,
                                       name='polynomial_decay_learning_rate')
    else:
      raise ValueError('learning_rate_decay_type [%s] was not recognized', learning_rate_decay_type)

  def get_init_fn(self, model, dump_dir, ctx=None):
    """Returns a function run by the chief worker to warm-start the training.

    Note that the init_fn is only run when initializing the model during the very
    first global step.

    Returns:
      An init function run by the supervisor.
    """
    # 1.step load from experiment
    if ctx is not None:
      if ctx.from_experiment is not None:
        no_activate_blocks = []
        for b in ctx.blocks():
          if not b.activate:
            no_activate_blocks.append(b.name)

        ablation_folder = '_'.join(sorted(no_activate_blocks))

        if len(ablation_folder) > 0:
          ctx.from_experiment = os.path.join(ctx.from_experiment, 'ablation', ablation_folder)
        # print(ctx.from_experiment)
        logger.info('load model from experiment %s' % ctx.from_experiment.split('/')[-2])
        latest_checkpoint = None
        try:
          latest_checkpoint = tf.train.latest_checkpoint(ctx.from_experiment)
        except:
          latest_checkpoint_index = -1
          for f in os.listdir(ctx.from_experiment):
            key_terms = f.split('.')
            if len(key_terms) >= 2 and key_terms[1] == 'ckpt':
              index = int(key_terms[0].split('_')[-1])
              if latest_checkpoint_index < index:
                latest_checkpoint_index = index
                latest_checkpoint = os.path.join(ctx.from_experiment, '.'.join(key_terms[0:2]))

        if latest_checkpoint is not None:
          variables_to_restore = {}
          # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
          model_variables = tf.global_variables()
          exclusions = []
          checkpoint_exclude_scopes = getattr(self, 'checkpoint_exclude_scopes', None)
          if checkpoint_exclude_scopes is not None:
            exclusions = [scope.strip()
                          for scope in checkpoint_exclude_scopes.split(',')]

            logger.info('exclude scope %s from the latest checkpoint' % (','.join(exclusions)))

          fuzzy_exclusions = []
          checkpoint_fuzzy_exclude_scopes = getattr(self, 'checkpoint_fuzzy_exclude_scopes', None)
          if checkpoint_fuzzy_exclude_scopes is not None:
            fuzzy_exclusions = [scope.strip()
                                for scope in checkpoint_fuzzy_exclude_scopes.split(',')]

            logger.info('exclude scope %s from the latest checkpoint' % (','.join(exclusions)))

          for var in model_variables:
            var_name = var.op.name

            excluded = False
            for exclusion in exclusions:
              if var_name.startswith(exclusion):
                excluded = True
                break

            for exclusion in fuzzy_exclusions:
              if exclusion in var_name:
                excluded = True
                break

            if excluded:
              continue

            variables_to_restore[var_name] = var

          return [slim.assign_from_checkpoint_fn(latest_checkpoint, variables_to_restore, ignore_missing_vars=True)]

    # Warn the user if a checkpoint exists in the train_dir. Then we'll be
    # ignoring the checkpoint anyway.
    # 2.step load from dump_dir
    if tf.train.latest_checkpoint(dump_dir):
      logger.info('Ignoring --checkpoint_path because a checkpoint already exists in %s' % dump_dir)
      # initilize model from dump_dir
      latest_checkpoint = tf.train.latest_checkpoint(dump_dir)
      variables_to_restore = {}
      # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
      model_variables = tf.global_variables()

      for var in model_variables:
        var_name = var.op.name
        variables_to_restore[var_name] = var

      logger.info('load model from %s' % latest_checkpoint)
      return [slim.assign_from_checkpoint_fn(latest_checkpoint, variables_to_restore)]

    # 3.step load from checkpoint_path
    checkpoint_path = getattr(self, 'checkpoint_path', None)
    if checkpoint_path is None:
      return None

    exclusions = []
    checkpoint_exclude_scopes = getattr(self, 'checkpoint_exclude_scopes', None)
    if checkpoint_exclude_scopes is not None:
      exclusions = [scope.strip()
                    for scope in checkpoint_exclude_scopes.split(',')]

    fuzzy_exclusions = []
    checkpoint_fuzzy_exclude_scopes = getattr(self, 'checkpoint_fuzzy_exclude_scopes', None)
    if checkpoint_fuzzy_exclude_scopes is not None:
      fuzzy_exclusions = [scope.strip()
                          for scope in checkpoint_fuzzy_exclude_scopes.split(',')]

    transfers = {}
    checkpoint_transfer_scopes = getattr(self, 'checkpoint_transfer_scopes', None)
    if checkpoint_transfer_scopes is not None:
      for scope in checkpoint_transfer_scopes.split(','):
        s_scope, t_scope = scope.split(':')
        transfers[s_scope.strip()] = t_scope.strip()

    # TODO(sguada) variables.filter_variables()
    auxilary_variables_to_restore = []
    variables_to_restore = {}
    # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
    model_variables = tf.global_variables()
    for var in model_variables:
      # transfer name
      var_name = var.op.name
      for k, v in transfers.items():
        if var.op.name.startswith(k):
          var_name = var_name.replace(k, v)

      # exclude name
      excluded = False
      for exclusion in exclusions:
        if var_name.startswith(exclusion):
          excluded = True
          break

      for exclusion in fuzzy_exclusions:
        if exclusion in var_name:
          excluded = True
          break

      # record
      if not excluded:
        if var_name not in variables_to_restore:
          variables_to_restore[var_name] = var
        else:
          is_ok = False
          for auxi_variables in auxilary_variables_to_restore:
            if var_name not in auxi_variables:
              auxi_variables[var_name] = var
              is_ok = True

          if not is_ok:
            auxilary_variables_to_restore.append({})
            auxilary_variables_to_restore[-1][var_name] = var

    auxilary_variables_to_restore.append(variables_to_restore)

    if tf.gfile.IsDirectory(checkpoint_path):
      checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)

    logger.info('load model from %s' % checkpoint_path)
    return [slim.assign_from_checkpoint_fn(checkpoint_path, vr, ignore_missing_vars=False) for vr in
            auxilary_variables_to_restore if len(vr) > 0]

  def __getattr__(self, item):
    if not item.endswith('_run'):
      return getattr(super(TFGANTrainer, self), item)

    def func(**kwargs):
      kwargs.update({'loss_name': item.replace('_run','')})
      return self.run(**kwargs)

    return func

  def restore_scopy_from(self, model, restore_scope, checkpoint_path):
    # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
    model_variables = tf.global_variables()
    variables_to_restore = {}
    for var in model_variables:
      if var.op.name.startswith(restore_scope):
        variables_to_restore[var.op.name] = var

    if len(variables_to_restore) == 0:
      return

    if tf.gfile.IsDirectory(checkpoint_path):
      checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)

    logger.info('restore %s scope from %s' % checkpoint_path)
    fn = slim.assign_from_checkpoint_fn(checkpoint_path, variables_to_restore, ignore_missing_vars=False)
    fn(self.sess)


  def run(self, *args, **kwargs):
    assert (len(args) <= 1)

    data_generator = None
    if len(args) == 1:
      data_generator = args[0]

    loss_name = None
    if 'loss_name' in kwargs:
      loss_name = kwargs['loss_name']
      kwargs.pop('loss_name')

    if data_generator is not None:
      return self._run_by_generator(data_generator, loss_name, **kwargs)
    else:
      return self._run_by_feed(loss_name, feed_dict=None, **kwargs)

  def _run_by_generator(self, data_generator, loss_name, **kwargs):
    # bind data
    with self.graph.as_default():
      feed_dict = {}
      if self._has_model_input and len(self.clones) > 1:
        logger.error('clones number > 1, must set different placeholder for every clone')
        exit(-1)

      if self._has_model_input:
        # generate data
        data = next(data_generator)

        for k, v in kwargs.items():
          if k not in self.cache:
            placeholder_tensor = self.graph.get_tensor_by_name('{}/{}:0'.format('input', k))
            self.cache[k] = placeholder_tensor

          feed_dict[self.cache[k]] = data[v] if (type(data) == tuple or type(data) == list) else data
      else:
        # set different placeholder for every clone
        for clone in self.clones:
          # generate data
          data = next(data_generator)

          for k, v in kwargs.items():
            if k not in self.cache:
              placeholder_tensor = None
              if len(self.clones) > 1:
                placeholder_tensor = self.graph.get_tensor_by_name('{}/{}:0'.format(clone[1][:-1], k))
              else:
                placeholder_tensor = self.graph.get_tensor_by_name('{}:0'.format(k))
              self.cache[k] = placeholder_tensor

            feed_dict[self.cache[k]] = data[v] if (type(data) == tuple or type(data) == list) else data

      return self._run_by_feed(loss_name, feed_dict=feed_dict, **kwargs)

  def _run_by_feed(self, loss_name=None, feed_dict=None, **kwargs):
    if feed_dict is None:
      feed_dict = {}
      for k_name, v_value in kwargs.items():
        if k_name not in self.cache:
          k_tensor = self.graph.get_tensor_by_name('{}:0'.format(k_name))
          self.cache[k_name] = k_tensor

        feed_dict[self.cache[k_name]] = v_value

    start_time = time.time()
    loss_name = loss_name if loss_name is not None else 'GAN'
    result = self.sess.run(self.trainop_list[loss_name],
                           feed_dict=feed_dict if len(feed_dict) > 0 else None)
    elapsed_time = int((time.time() - start_time) * 100) / 100.0

    if self.ctx.recorder is not None and self.ctx.recorder.model_fn is not None:
      self.ctx.recorder.action(result[-1])
      result = result[:-1]

    # record iterator count
    self.iter_at += 1
    if loss_name not in self.log_iter_at:
      self.log_iter_at[loss_name] = 0
    self.log_iter_at[loss_name] += 1

    if self.summary_op is not None:
      summary_op_val = result[0]
      result = result[1:]
      self.train_writer.add_summary(summary_op_val, self.iter_at)

    # record elapsed time
    self.time_stat.add(elapsed_time)

    # print log
    if self.is_training:
      loss_val = 0.0
      if type(result) == list and len(result) >= 2:
        loss_val = result[1]
      elif type(result) == list and len(result) == 1:
        loss_val = result[0]
      else:
        loss_val = result

      # record loss value
      self.loss_log[loss_name].add(loss_val)

      #
      if self.log_iter_at[loss_name] % self.log_every_n_steps == 0:
        if not self.is_distribute_training or (self.is_distribute_training and self.rank == 0):
          loss_log_str = ''
          for k,v in self.loss_log.items():
            loss_log_str = loss_log_str + ' %s %f'%(k, v.get())

          logger.info('(PID: %s) INFO: %s lr %f at iterator %d (%f sec/step)' %
                      (str(os.getpid()),
                       loss_log_str,
                       self.sess.run(self.lr_list[loss_name]),
                       self.log_iter_at[loss_name],
                       float(self.time_stat.get())))
    else:
      if not self.is_distribute_training or (self.is_distribute_training and self.rank == 0):
        logger.info('(PID: %s) INFO: (%f sec/step)' % (str(os.getpid()), float(self.time_stat.get())))

    return result[0] if type(result) == list and len(result) == 1 else result

  # 3.step snapshot running state
  def snapshot(self, epoch=0, iter=-1):
    logger.info('snapshot at %d in %d epoch' % (self.iter_at, epoch))
    if not os.path.exists(self.dump_dir):
      os.makedirs(self.dump_dir)

    model_filename = "{prefix}_{infix}_{d}_{e}.ckpt".format(prefix=self.snapshot_prefix,
                                                        infix=self.snapshot_infix,
                                                        d=self.iter_at if iter < 0 else iter,
                                                        e=epoch)
    model_filepath = os.path.join(self.dump_dir, model_filename)

    # save checkpoint
    self.saver.save(self.sess, model_filepath)

  def training_deploy(self, model, *args,**kwargs):
    tf.logging.set_verbosity(tf.logging.INFO)
    with tf.Graph().as_default() as graph:
      # default graph
      self.graph = graph

      # session
      config = tf.ConfigProto(allow_soft_placement=True)
      config.gpu_options.allow_growth = True
      devices = self.ctx.devices if len(self.ctx.devices) > 0 else self.devices
      config.gpu_options.visible_device_list = ','.join(str(x) for x in devices) if len(devices) > 0 else ''
      self.sess = tf.Session(graph=graph, config=config)

      #######################
      # Config model deploy #
      #######################
      deploy_config = tfmodel_deploy.DeploymentConfig(num_clones=1,
                                                      devices=self.devices,
                                                      clone_on_cpu=self.clone_on_cpu,
                                                      replica_id=self.replica_id,
                                                      num_replicas=self.worker_replicas,
                                                      num_ps_tasks=self.num_ps_tasks)

      # init some info
      with tf.device(deploy_config.inputs_device()):
        #############################
        #### Define model input #####
        #############################
        with tf.variable_scope('input'):
          data_queue = self.ctx.model.model_input(self.is_training)
          if data_queue is not None:
            self._has_model_input = True

        func = model.model_fn
        @functools.wraps(func)
        def network_fn(*args, **kwargs):
          #
          logger.info('building computing graph')
          res = func(self.is_training, *args, **kwargs)
          tf.train.write_graph(self.sess.graph_def, self.dump_dir, 'graph.pbtxt')
          return res

        ####################################
        ####### Create summary      ########
        ####################################
        summaries = set(tf.get_collection(tf.GraphKeys.SUMMARIES))

        #############################
        # Create model clones #######
        #############################
        self.clones = tfmodel_deploy.create_clones(deploy_config,
                                                   network_fn,
                                                   [data_queue] if data_queue is not None else None)

        #########################################
        # Configure the optimization procedure. #
        #########################################
        with tf.device(deploy_config.optimizer_device()):
          num_samples = self.num_samples if self.num_samples > 0 else self.ctx.data_source.size
          trainable_variables = self.get_variables_to_train()       # all
          update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)   # all

          for loss in tf.get_collection(tf.GraphKeys.LOSSES, deploy_config.clone_scope(0)):
            summaries.add(tf.summary.scalar('losses/%s' % loss.op.name, loss))

          for loss_name, loss_config in kwargs.items():
            loss_scope = loss_config['scope']

            # config loss log
            self.loss_log[loss_name] = MovingAverage(10)

            # Extract loss variable
            self.loss_list[loss_name] = graph.get_tensor_by_name('{}:0'.format(loss_name))

            if 'learning_rate' in loss_config:
              self.lr_list[loss_name] = graph.get_tensor_by_name('{}:0'.format(loss_config['learning_rate']))
              self.global_step_list[loss_name] = None
            else:
              global_step = tf.Variable(0, trainable=False)
              self.lr_list[loss_name] = self.configure_learning_rate(num_samples, global_step)
              self.global_step_list[loss_name] = global_step

            summaries.add(tf.summary.scalar('%s_learning_rate'%loss_name, self.lr_list[loss_name]))

            # config optimization procedure
            self.optimizer_list[loss_name] = self.configure_optimizer(self.lr_list[loss_name])

            # config related variables
            for loss_scope_name in loss_scope.split(','):
              if loss_name not in self.var_list:
                self.var_list[loss_name] = []
              self.var_list[loss_name].extend([var for var in trainable_variables if loss_scope_name in var.name])

            # get update ops
            for loss_scope_name in loss_scope.split(','):
              if loss_name not in self.update_list:
                self.update_list[loss_name] = []
              self.update_list[loss_name].extend([var for var in update_ops if loss_scope_name in var.name])


        summaries |= set(tf.get_collection(tf.GraphKeys.SUMMARIES, deploy_config.clone_scope(0)))

        # Merge all summaries together.
        self.summary_op = tf.summary.merge(list(summaries), name='summary_op')

        # Variables to train.
        for loss_name, loss_var in self.loss_list.items():
          optimizer = self.optimizer_list[loss_name]

          with tf.name_scope(self.clones[0].scope):
            with tf.device(self.clones[0].device):
              clone_grad = optimizer.compute_gradients(loss_var, var_list=self.var_list[loss_name])
              grad_update = optimizer.apply_gradients(clone_grad,global_step=self.global_step_list[loss_name] if self.global_step_list[loss_name] is not None else None)

              self.update_list[loss_name].append(grad_update)

              with tf.control_dependencies([tf.group(*self.update_list[loss_name])]):
                train_op = tf.identity(loss_var, name='train_op_%s'%loss_name)
                self.trainop_list[loss_name] = train_op

                if self.clones[0].outputs is not None:
                  if type(self.clones[0].outputs) == dict:
                    for k,v in self.clones[0].outputs.items():
                      if type(k) != str:
                        k = k.name.replace(':0','')

                      if k == loss_name:
                        self.trainop_list[loss_name] = [train_op]
                        if type(v) == list or type(v) == tuple:
                          self.trainop_list[loss_name].extend(list(v))
                        else:
                          self.trainop_list[loss_name].append(v)

                if type(self.trainop_list[loss_name]) != list:
                  self.trainop_list[loss_name] = [self.trainop_list[loss_name]]

                if self.summary_op is not None:
                  val_ops_temp = [self.summary_op]
                  val_ops_temp.extend(self.trainop_list[loss_name])
                  self.trainop_list[loss_name] = val_ops_temp

        # summary write
        if not os.path.exists(os.path.join(self.dump_dir, 'summary')):
          os.makedirs(os.path.join(self.dump_dir, 'summary'))

        self.train_writer = tf.summary.FileWriter(os.path.join(self.dump_dir, 'summary'), graph)

        # Global initialization
        self.sess.run(tf.global_variables_initializer())
        self.sess.run(tf.local_variables_initializer())
        # coord
        self.coord = tf.train.Coordinator()
        self.threads = tf.train.start_queue_runners(sess=self.sess, coord=self.coord)

        custom_dataset_queue = tf.get_collection('CUSTOM_DATASET_QUEUE')
        if len(custom_dataset_queue) > 0:
          custom_dataset_queue[0].coord = self.coord
          custom_threads = custom_dataset_queue[0].start_threads(self.sess)
          self.threads.extend(custom_threads)

        # Training saver
        # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
        self.saver = tf.train.Saver(max_to_keep=2)

        # Restore from checkpoint
        restore_fns = self.get_init_fn(model, self.dump_dir, self.ctx)
        if restore_fns is not None:
          for restore_fn in restore_fns:
            restore_fn(self.sess)

        # resotre from auxilary checkpoint
        for auxilary_scope, auxilary_checkpoint in self.auxilary_checkpoints.items():
          self.restore_scopy_from(model, auxilary_scope, auxilary_checkpoint)

  def infer_deploy(self, model, *args, **kwargs):
    tf.logging.set_verbosity(tf.logging.INFO)
    with tf.Graph().as_default() as graph:
      # Default graph
      self.graph = graph
      # Session
      config = tf.ConfigProto(allow_soft_placement=True)
      config.gpu_options.allow_growth = True
      devices = self.ctx.devices if len(self.ctx.devices) > 0 else self.devices
      config.gpu_options.visible_device_list = ','.join(str(x) for x in devices) if len(devices) > 0 else ''
      self.sess = tf.Session(graph=graph, config=config)

      #######################
      # Config model_deploy #
      #######################
      deploy_config = tfmodel_deploy.DeploymentConfig(num_clones=1,
                                                      devices=self.devices,
                                                      clone_on_cpu=getattr(self, 'clone_on_cpu', False),
                                                      replica_id=0,
                                                      num_replicas=1,
                                                      num_ps_tasks=0)

      # init some info
      with tf.device(deploy_config.inputs_device()):
        #############################
        ####    define model input ##
        #############################
        data_queue = None
        if self.ctx.ant is not None:
          with tf.variable_scope('input'):
            data_queue = self.ctx.model.model_input(self.is_training)
            if data_queue is not None:
              self._has_model_input = True

        #############################
        ####    define model       ##
        #############################
        func = model.model_fn
        @functools.wraps(func)
        def network_fn(*args, **kwargs):
          res = func(self.is_training, *args, **kwargs)
          return res

        #######################
        # Create model clones #
        #######################
        self.clones = tfmodel_deploy.create_clones(deploy_config, network_fn, [data_queue] if data_queue is not None else None)

        # write graph
        tf.train.write_graph(graph.as_graph_def(), self.dump_dir, 'infer_graph.pbtxt')

        # Global initialization
        self.sess.run(tf.global_variables_initializer())
        self.sess.run(tf.local_variables_initializer())

        self.coord = tf.train.Coordinator()
        self.threads = tf.train.start_queue_runners(sess=self.sess, coord=self.coord)

        custom_dataset_queue = tf.get_collection('CUSTOM_DATASET_QUEUE')
        if len(custom_dataset_queue) > 0:
          custom_dataset_queue[0].coord = self.coord
          custom_threads = custom_dataset_queue[0].start_threads(self.sess)
          self.threads.extend(custom_threads)

        # Restore from checkpoint
        restore_fns = self.get_init_fn(model, self.dump_dir, self.ctx)
        if restore_fns is not None:
          for restore_fn in restore_fns:
            restore_fn(self.sess)

        # model saver
        # model_variables = slim.get_model_variables() if model.model_variables is None else model.model_variables
        self.saver = tf.train.Saver(max_to_keep=1)

        # snapshot
        self.snapshot(0)

        # Value ops
        self.trainop_list['GAN'] = self.clones[0].outputs
        if type(self.trainop_list['GAN']) != list and type(self.trainop_list['GAN']) != tuple:
          self.trainop_list['GAN'] = self.trainop_list['GAN']
        if type(self.trainop_list["GAN"]) == tuple:
          self.trainop_list['GAN'] = list(self.trainop_list['GAN'])

        # Append recorder model fn
        if self.ctx.recorder is not None and self.ctx.recorder.model_fn is not None:
          self.trainop_list['GAN'].append(self.ctx.recorder.model_fn)

  def deploy(self, model, *args, **kwargs):
    # model context
    self.ctx.model = model
    model.trainer = self

    if self.is_training:
      self.training_deploy(model, *args, **kwargs)
    else:
      self.infer_deploy(model, *args, **kwargs)

      if self.ctx.ant is None:
        logger.info('successfully deploy model')
        exit(0)

  def wait_until_clear(self):
    if self.coord is not None:
      self.coord.request_stop()
      self.coord.join(self.threads)


# encoding=utf-8
# @Time    : 17-5-8
# @File    : cpu.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import re
import psutil
import os
from antgo.utils import timer
import numpy.distutils.cpuinfo
unit = {'b':1, 'k':2**10, 'm':2**20, 'g':2**30}


def get_nr_cpu():
    return psutil.cpu_count()


def cpu_running_info(pid=None,interval=1):
    '''
    key:
        mem_usage is RSS memory.RSS is thetotal memory actually held in RAM for a process.
        RSS can be misleading, because it reports the total all of the shared libraries 
        that the process uses, even though a shared library is only loaded into memory 
        once regardless of how many processes use it.
    :param pid: 
    :return: 
    '''
    mem_max = psutil.virtual_memory().total / unit['m']

    cpus = []
    cpu_info = numpy.distutils.cpuinfo.cpu.info
    if type(cpu_info) == list:
        for cpu_index in range(len(cpu_info)):
            if 'model name' in cpu_info[cpu_index]:
                cpus.append(cpu_info[cpu_index]['model name'])
            else:
                cpus.append('unkown')
    else:
        if 'model name' in cpu_info:
            cpus.append(cpu_info['model name'])
        else:
            cpus.append('unkown')

    if pid is not None:
        p = psutil.Process(pid)
        mem_usage = p.memory_percent() * 0.01 * mem_max
        cpu_usage = p.cpu_percent(interval=interval)
        num_threads = p.num_threads()
        if getattr(p,'cpu_affinity',None) is not None:
            occupy_cpus = p.cpu_affinity()
        else:
            occupy_cpus = [0]

        return {'cpus': cpus,'cpu_mem_usage':mem_usage,
                'cpu_mem_max':mem_max,'cpu_util':cpu_usage,
                'occupy_cpus':occupy_cpus,'num_threads':num_threads}

    return {'cpus': cpus,'cpu_mem_max':mem_max}


# -*- coding: UTF-8 -*-
# File: omniglot.py
# Author: jian(jian@mltalker.com)
from __future__ import division
from __future__ import unicode_literals

import numpy as np
import os
import random
import copy
import time
from antgo.dataflow.dataset import Dataset
from antgo.utils.fs import download
from antgo.utils.fs import maybe_here_fixed_3_hierarchy

__all__ = ['Omniglot']
OmniglotURL = 'https://github.com/brendenlake/omniglot/raw/master/python/'
class Omniglot(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(Omniglot, self).__init__(train_or_test, dir)
    assert(train_or_test in ['train', 'val', 'sample'])
    self.train_or_test = train_or_test
    self.data_folder = dir       # data folder

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    maybe_data_dir = maybe_here_fixed_3_hierarchy(os.path.join(self.data_folder, train_or_test), '.png')
    if maybe_data_dir is None:
      if train_or_test == 'train':
        self.download(os.path.join(self.data_folder, train_or_test), ['images_background.zip'], OmniglotURL, auto_unzip=True)
        maybe_data_dir = os.path.join(self.data_folder, train_or_test, 'images_background')
      else:
        self.download(os.path.join(self.data_folder, train_or_test), ['images_evaluation.zip'], OmniglotURL, auto_unzip=True)
        maybe_data_dir = os.path.join(self.data_folder, train_or_test, 'images_evaluation')

    self.character_folders = [os.path.join(maybe_data_dir, family, character) \
                              for family in os.listdir(maybe_data_dir) \
                              if os.path.isdir(os.path.join(maybe_data_dir, family)) \
                              for character in os.listdir(os.path.join(maybe_data_dir, family))]

    self.character_labels = range(len(self.character_folders))
    self.labels_num = len(self.character_labels)

    self.character_image_paths = [
      [os.path.join(self.character_folders[i], p) for p in os.listdir(self.character_folders[i])] for i in
      range(len(self.character_folders))]

    self.samples_label = []
    self.samples_index = []
    for label in range(self.labels_num):
      for i in range(len(self.character_image_paths[label])):
        self.samples_label.append(label)
        self.samples_index.append(i)

    self.samples_num = 0
    for ll in self.character_image_paths:
      self.samples_num += len(ll)

    self.ids = list(range(self.samples_num))
    self.samples = self.make_data(self.character_image_paths, self.character_labels)

    # fixed seed
    self.seed = time.time()

  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')
    assert (split_method == 'holdout')

    val_omniglot = Omniglot('val', self.dir)
    return self, val_omniglot

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = copy.deepcopy(self.ids)
      if self.rng:
          self.rng.shuffle(idxs)

      # generate single image
      for k in idxs:
          yield [self.load_image(self.samples[k][2]), self.samples[k][1]]

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    return [self.load_image(self.samples[self.ids[id]][2]), self.samples[self.ids[id]][1]]

  @property
  def size(self):
      return len(self.ids)


from setuptools import setup
from Cython.Build import cythonize
from distutils.extension import Extension
import os
import numpy as np
import shutil

# install: python setup.py build_ext install -r requirements.txt (from github)
ext_modules = [
    Extension('antgo.utils._mask',
              sources=['antgo/cutils/maskApi.c', 'antgo/utils/_mask.pyx'],
              include_dirs = [np.get_include(), 'antgo/cutils'],
              extra_compile_args=['-Wno-cpp', '-Wno-unused-function', '-std=c99'],
    ),
    Extension('antgo.utils._bbox',
              ["antgo/utils/_bbox.pyx"],
              include_dirs=[np.get_include()],
              extra_compile_args=["-Wno-cpp", "-Wno-unused-function", '-std=c99']
    ),
    Extension('antgo.utils._nms',
              ["antgo/utils/_nms.pyx"],
              include_dirs=[np.get_include()],
              extra_compile_args=["-Wno-cpp", "-Wno-unused-function", '-std=c99']
              ),
    Extension('antgo.utils._resize',
              ["antgo/utils/_resize.pyx"],
              include_dirs=[np.get_include()],
              extra_compile_args=["-Wno-cpp", "-Wno-unused-function"]
              ),
]

def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='antgo',
      version='0.0.7',
      description='machine learning experiment platform',
      __short_description__='machine learning experiment platform',
      url='https://github.com/jianzfb/antgo',
      author='jian',
      author_email='jian@mltalker.com',
      packages=['antgo',
                'antgo.ant',
                'antgo.utils',
                'antgo.utils.shared_queue',
                'antgo.dataflow',
                'antgo.dataflow.dataset',
                'antgo.dataflow.datasynth',
                'antgo.dataflow.imgaug',
                'antgo.resource',
                'antgo.resource.templates',
                'antgo.resource.static',
                'antgo.resource.browser',
                'antgo.resource.browser.static',
                'antgo.resource.browser.static.css',
                'antgo.resource.browser.static.fonts',
                'antgo.resource.browser.static.img',
                'antgo.resource.browser.static.js',
                'antgo.resource.batch',
                'antgo.resource.batch.static',
                'antgo.resource.batch.static.css',
                'antgo.resource.batch.static.fonts',
                'antgo.resource.batch.static.img',
                'antgo.resource.batch.static.js',
                'antgo.crowdsource',
                'antgo.measures',
                'antgo.task',
                'antgo.trainer',
                'antgo.activelearning',
                'antgo.activelearning.samplingmethods',
                'antgo.cutils',
                'antgo.annotation',
                'antgo.sandbox',
                'antgo.codebook',
                'antgo.codebook.tf'],
      ext_modules=cythonize(ext_modules),
      entry_points={'console_scripts': ['antgo=antgo.main:main'], },
      long_description=readme(),
      include_package_data=True,
      zip_safe=False,)



"""Tools for visualizing a lusmu graph

Copyright 2013 Eniram Ltd. See the LICENSE file at the top-level directory of
this distribution and at https://github.com/akaihola/lusmu/blob/master/LICENSE

"""

# pylint: disable=W0212
#         Allow access to protected members of client classes
# pylint: disable=W0142
#         Allow * and ** magic

from __future__ import print_function, unicode_literals

import re
import subprocess
from textwrap import dedent
import collections
from antgo.utils.serialize import *


class Link(collections.namedtuple('Link', ['label', 'args'])):
  """
  """


class NodeVis(object):
  def __init__(self, **kwargs):
    self.name = kwargs['name']                                      # name
    self.op_name = kwargs['op_name'].replace('/', '<br />')         # op name
    self.link_nodes = [] if 'link_nodes' not in kwargs else kwargs['link_nodes']
    self.label = '' if 'label' not in kwargs else kwargs['label']
    self.id = kwargs['id']
    self._in_degree = 0 if '_in_degree' not in kwargs else kwargs['_in_degree']
    self._out_degree = 0 if '_out_degree' not in kwargs else kwargs['_out_degree']
    self._attr = {} if '_attr' not in kwargs else kwargs['_attr']
    self.tag = {} if 'tag' not in kwargs else kwargs['tag']

  def add_link(self, node, link):
    for ln in self.link_nodes:
      if node == ln['node']:
        return False
    
    self.link_nodes.append({'node': node, 'link': link})
    return True

  @property
  def linked_node(self):
    for a in self.link_nodes:
      yield a

  @property
  def in_degree(self):
    return self._in_degree

  def in_degree_increment(self):
    self._in_degree += 1

  @property
  def out_degree(self):
    return self._out_degree

  def out_degree_increment(self):
    self._out_degree += 1


class GraphVis(object):
  def __init__(self, **kwargs):
    self._name = kwargs['name'] if 'name' in kwargs else kwargs['_name']
    self._graph_nodes = {} if '_graph_nodes' not in kwargs else kwargs['_graph_nodes']
    self._graph_nodes_indegree = {} if '_graph_nodes_indegree' not in kwargs else kwargs['_graph_nodes_indegree']
    self._graph_nodes_outdegree = {} if '_graph_nodes_outdegree' not in kwargs else kwargs['_graph_nodes_outdegree']
    self._nodes_num = 0 if '_nodes_num' not in kwargs else kwargs['_nodes_num']

  def add_node(self, name, op_name, label, tag=''):
    if name not in self._graph_nodes:
      self._graph_nodes[name] = NodeVis(name=name, op_name=op_name, id=self._nodes_num, label=label, tag=tag)
      self._graph_nodes_indegree[name] = 0
      self._graph_nodes_outdegree[name] = 0
      self._nodes_num += 1

  def add_link(self, from_node, to_node, link):
    try:
      assert(from_node in self._graph_nodes)
      assert(to_node in self._graph_nodes)
      
      flag = self._graph_nodes[from_node].add_link(self._graph_nodes[to_node], link)
      if flag:
        # add new link successfully
        self._graph_nodes[from_node].out_degree_increment()
        self._graph_nodes[to_node].in_degree_increment()
    except:
      print(from_node)
      print(to_node)

  @property
  def nodes(self):
    for k,v in self._graph_nodes.items():
      yield v

  def node(self, name):
    return self._graph_nodes[name]


def node_format(node_id, node):
  # shape = 'oval' if isinstance(node, Input) else 'box'
  shape='circle'
  # action = ('{br}{br}<FONT COLOR="#888888">{action}</FONT>'
  #           .format(br=' <BR ALIGN="LEFT"/>',
  #                   action=get_action_name(node._action))
  #           if isinstance(node, Node)
  #           else '')
  action = ''
  if node.in_degree == 0:
    yield ('  {node_id} '
           '[label=<<B>{name}</B>'
           '{action}>'
           ' shape={shape} style=filled fillcolor=forestgreen fontcolor=white penwidth=1];'
           .format(node_id=node_id,
                   name=node.op_name.replace(':', ':<BR ALIGN="CENTER"/>'),
                   action=action,
                   shape=shape))
  elif node.out_degree == 0:
    yield ('  {node_id} '
           '[label=<<B>{name}</B>'
           '{action}>'
           ' shape={shape} style=filled fillcolor=mediumpurple fontcolor=white penwidth=1];'
           .format(node_id=node_id,
                   name=node.op_name.replace(':', ':<BR ALIGN="CENTER"/>'),
                   action=action,
                   shape=shape))
  else:
    yield ('  {node_id} '
           '[label=<<B>{name}</B>'
           '{action}>'
           ' shape={shape} style=filled fillcolor=goldenrod fontcolor=white penwidth=1];'
           .format(node_id=node_id,
                   name=node.op_name.replace(':', ':<BR ALIGN="CENTER"/>'),
                   action=action,
                   shape=shape))
  yield '  edge [color=darkgrey ];'


def graphviz_net(graph, format_node, who_is_input):
  """Generate source lines for a Graphviz graph definition"""
  all_nodes = sorted(graph.nodes, key=id)
  if who_is_input is not None and len(who_is_input) > 0:
    input_nodes = [n for n in all_nodes if n.name in who_is_input]
  else:
    input_nodes = [n for n in all_nodes if n.in_degree==0 and n.out_degree > 0]

  yield 'digraph gr {'
  yield '  graph [ dpi = 12 ];'
  yield '  rankdir = TB;'

  yield '  { rank = source;'
  for node in input_nodes:
    yield '    n{};'.format(node.id)
  yield '  }'

  block_names = {'regular': []}
  cell_names = {'regular': []}
  block_cell_map = {}
  for node in all_nodes:
    if 'block' in node.tag and node.tag['block'] != '':
      if node.tag['block'] not in block_names:
        block_names[node.tag['block']] = []

      block_names[node.tag['block']].append(node)

      if 'cell' in node.tag and node.tag['cell'] != '':
        if node.tag['cell'] not in cell_names:
          cell_names[node.tag['cell']] = []
        cell_names[node.tag['cell']].append(node)

        if node.tag['block'] not in block_cell_map:
          block_cell_map[node.tag['block']] = []

        if node.tag['cell'] not in block_cell_map[node.tag['block']]:
          block_cell_map[node.tag['block']].append(node.tag['cell'])
      else:
        cell_names['regular'].append(node)
    else:
      block_names['regular'].append(node)

  for k,v in block_names.items():
    if k == 'regular':
      for node in v:
        if node.in_degree == 0 and node.out_degree == 0:
          continue

        # support block tag
        for line in format_node('n{}'.format(node.id), node):
          yield line
    else:
      yield ' subgraph cluster_%s{'%k
      yield ' label="%s";'%k
      yield ' bgcolor="mintcream";'

      if len(block_cell_map) == 0:
        for node in v:
          if node.in_degree == 0 and node.out_degree == 0:
            continue

          for line in format_node('n{}'.format(node.id), node):
            yield line
      else:
        for cell_name in block_cell_map[k]:
          yield ' subgraph cluster_%s{' % cell_name
          yield ' label="%s";' % cell_name
          yield ' bgcolor="cornflowerblue";'
          for node in cell_names[cell_name]:
            if node.in_degree == 0 and node.out_degree == 0:
              continue

            for line in format_node('n{}'.format(node.id), node):
              yield line
          yield ' }'
        pass
      yield ' }'

  for node in all_nodes:
    # if node.in_degree == 0 and node.out_degree == 0:
    #   continue
    #
    # # support block tag
    # for line in format_node('n{}'.format(node.id), node):
    #   yield line

    for other in node.linked_node:
      if other['link'][0] != '':
        yield ('  n{node} -> n{other}[label={label},fontcolor=darkgreen,penwidth=2,arrowsize=0.5];'
             .format(node=node.id, other=other['node'].id, label=other['link'][0]))
      else:
        yield ('  n{node} -> n{other}[fontcolor=darkgreen,penwidth=2,arrowsize=0.5];'
             .format(node=node.id, other=other['node'].id))
  yield '}'


def graphviz_net_svg(graph, file_path, who_is_input=[]):
  # 1.step extract export format
  image_format = file_path.split('.')[-1].lower()

  # 2.step dot
  graphviz = subprocess.Popen(['dot',
                                 '-T{}'.format(image_format),
                                 '-Gdpi=50',
                                 '-o', file_path],
                                stdin=subprocess.PIPE)


  # 3.step generate dot code
  source = '\n'.join(graphviz_net(graph, node_format, who_is_input))

  if image_format == 'svg':
    source = re.sub(r'(<svg\s[^>]*>)',
                 dedent(r'''
      \1
      <style type="text/css"><![CDATA[
          g.node:hover {
              stroke-width: 2;
          }
      ]]></style>
      '''),
                 source,
                 re.S)
  
  # 4.step save to file
  graphviz.communicate(source.encode('utf-8'))

  # 5.step svg content
  if image_format == 'svg':
    fp = open(file_path, 'r')
    content = fp.read()
    fp.close()
    return content

  return None

# test_graph = Graph(name='testnet')
# test_graph.add_node(name='conv', label='hello')
# test_graph.add_node(name='pool', label='world')
# test_graph.add_node(name='input', label='www')
# test_graph.add_node(name='ssd', label='hhh')
# test_graph.add_link('conv','pool',Link('ab',''))
# test_graph.add_link('input','conv',Link('bd',''))
# test_graph.add_link('input','ssd',Link('ac',''))
#
# ss = Encoder().encode(test_graph)
# print(ss)
# dd = Decoder().decode(ss)
# print(ss)

# graph_content = graph_net_visualization(test_graph, '/Users/zhangken/Downloads/testnet.svg')
# print(graph_content)
  # net_visualization(graph, '/Users/zhangken/Downloads/11.svg',net_node_format)


# -*- coding: UTF-8 -*-
# @Time    : 2019-04-30 18:32
# @File    : simplevideos.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset.dataset import *
import os
import copy
from PIL import Image
import imageio

__all__ = ['SimpleVideos']


class SimpleVideos(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(SimpleVideos, self).__init__(train_or_test, dir, params)

    self.data_files = []
    for data_file in os.listdir(os.path.join(self.dir, self.train_or_test)):
      if data_file.lower().split('.')[-1] in ['mp4', 'mov', 'avi']:
        file_path = os.path.join(self.dir, self.train_or_test, data_file)
        self.data_files.append(file_path)

    assert (len(self.data_files) > 0)
    self.ids = [i for i in range(len(self.data_files))]

  def _video_iterator(self, video_path):
    reader = imageio.get_reader(video_path)
    for im in reader:
      img_data = np.fromstring(im.tobytes(), dtype=np.uint8)
      img_data = img_data.reshape((im.shape[0], im.shape[1], im.shape[2]))
      yield img_data, None

  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      ids = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(ids)

      for id in ids:
        video_path = self.data_files[id]
        yield self._video_iterator(video_path), None

  def at(self, id):
    raise NotImplementedError

  @property
  def size(self):
    return len(self.ids)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')

    return self, SimpleVideos('val', self.dir, self.ext_params)



# -*- coding: UTF-8 -*-
# @Time    : 18-1-3
# @File    : dataset.py
# @Author  : Jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import threading
import time
import numpy as np
import tensorflow as tf
from antgo.utils import logger


class TFQueueDataset(object):
  def __init__(self,
               datasource,
               dtype=[tf.uint8],
               shape=[[None, None, None]],
               max_queue_size=32,
               wait_time=0.01):
    # Change the shape of the input data here with the parameter shapes.
    self.wait_time = wait_time
    self.max_queue_size = max_queue_size
    self.dtype_list = dtype
    self.shape_list = shape

    self.threads = []
    self._coord = None
    self.sample_placeholder = []
    self.datasource = datasource
    self._size = None
  
  @property
  def size(self):
    if self._size is None:
      return self.datasource.size
    else:
      return self._size

  @size.setter
  def size(self, val):
    self._size = val

  @property
  def coord(self):
    return self._coord

  @coord.setter
  def coord(self, val):
    self._coord = val
    
  def dequeue(self):
    return self.model_fn()

  def model_fn(self):
    self.queue = tf.PaddingFIFOQueue(self.max_queue_size, self.dtype_list, shapes=self.shape_list)
    self.queue_size = self.queue.size()
    for i in range(len(self.dtype_list)):
      self.sample_placeholder.append(tf.placeholder(dtype=self.dtype_list[i], shape=None))
    self.enqueue = self.queue.enqueue(self.sample_placeholder)
    tf.add_to_collection('CUSTOM_DATASET_QUEUE', self)
    
    return self.queue.dequeue()
  
  def thread_main(self, sess):
    stop = False
    while not stop:
      self.datasource._reset_iteration_state()
      iterator = self.datasource.iterator_value()
      
      for data in iterator:
        try:
          while self.queue_size.eval(session=sess) == self.max_queue_size:
            if self.coord.should_stop():
              self.queue.close()
              break
            time.sleep(self.wait_time)

          if self.coord.should_stop():
            stop = True
            self.queue.close()
            break
          feed_dict = {}
          if type(data) == list or type(data) == tuple:
            for i in range(len(data)):
              feed_dict.update({self.sample_placeholder[i]: data[i]})
          else:
            feed_dict = {self.sample_placeholder[0]: data}
          sess.run(self.enqueue, feed_dict=feed_dict)
        except:
          logger.error('couldnt feed data into tensorflow pipeline')
          pass
        
  def start_threads(self, sess):
    for _ in range(1):
      thread = threading.Thread(target=self.thread_main, args=(sess,))
      thread.daemon = True  # Thread will close when parent quits.
      thread.start()
      self.threads.append(thread)
    return self.threads


# -*- coding: UTF-8 -*-
# File: logger.py
# Author: jian<jian@mltalker.com>
from __future__ import unicode_literals

import logging
import os, shutil
import os.path
from termcolor import colored
from datetime import datetime
from six.moves import input
import sys

from .utils import memoized

__all__ = []

class _MyFormatter(logging.Formatter):
    def format(self, record):
        date = colored('[%(asctime)s @%(filename)s:%(lineno)d]', 'green')
        msg = '%(message)s'
        if record.levelno == logging.WARNING:
            fmt = date + ' ' + colored('WRN', 'red', attrs=['blink']) + ' ' + msg
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            fmt = date + ' ' + colored('ERR', 'red', attrs=['blink', 'underline']) + ' ' + msg
        else:
            fmt = date + ' ' + msg
        if hasattr(self, '_style'):
            # Python3 compatibilty
            self._style._fmt = fmt
        self._fmt = fmt
        return super(_MyFormatter, self).format(record)

def _getlogger():
    logger = logging.getLogger('tensorpack')
    logger.propagate = False
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_MyFormatter(datefmt='%m%d %H:%M:%S'))
    logger.addHandler(handler)
    return logger
_logger = _getlogger()


def get_time_str():
    return datetime.now().strftime('%m%d-%H%M%S')
# logger file and directory:
global LOG_FILE, LOG_DIR
LOG_DIR = None
def _set_file(path):
    if os.path.isfile(path):
        backup_name = path + '.' + get_time_str()
        shutil.move(path, backup_name)
        info("Log file '{}' backuped to '{}'".format(path, backup_name))
    hdl = logging.FileHandler(
        filename=path, encoding='utf-8', mode='w')
    hdl.setFormatter(_MyFormatter(datefmt='%m%d %H:%M:%S'))
    _logger.addHandler(hdl)

def set_logger_dir(dirname, action=None):
    """
    Set the directory for global logging.
    :param dirname: log directory
    :param action: an action (k/b/d/n) to be performed. Will ask user by default.
    """
    global LOG_FILE, LOG_DIR
    if os.path.isdir(dirname):
        if not action:
            _logger.warn("""\
Directory {} exists! Please either backup/delete it, or use a new directory. \
If you're resuming from a previous run you can choose to keep it.""".format(dirname))
            _logger.info("Select Action: k (keep) / b (backup) / d (delete) / n (new):")
        while not action:
            action = input().lower().strip()
        act = action
        if act == 'b':
            backup_name = dirname + get_time_str()
            shutil.move(dirname, backup_name)
            info("Directory '{}' backuped to '{}'".format(dirname, backup_name))
        elif act == 'd':
            shutil.rmtree(dirname)
        elif act == 'n':
            dirname = dirname + get_time_str()
            info("Use a new log directory {}".format(dirname))
        elif act == 'k':
            pass
        else:
            raise ValueError("Unknown action: {}".format(act))
    LOG_DIR = dirname
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != 17:
            raise e
    LOG_FILE = os.path.join(dirname, 'log.log')
    _set_file(LOG_FILE)

# export logger functions
for func in ['info', 'warning', 'error', 'critical', 'warn', 'exception', 'debug']:
    locals()[func] = getattr(_logger, func)

def disable_logger():
    """ disable all logging ability from this moment"""
    for func in ['info', 'warning', 'error', 'critical', 'warn', 'exception', 'debug']:
        globals()[func] = lambda x: None

def auto_set_dir(action=None, overwrite_setting=False):
    """ set log directory to a subdir inside 'train_log', with the name being
    the main python file currently running"""
    if LOG_DIR is not None and not overwrite_setting:
        return
    mod = sys.modules['__main__']
    basename = os.path.basename(mod.__file__)
    set_logger_dir(
            os.path.join('train_log',
                basename[:basename.rfind('.')]),
            action=action)



# -*- coding: UTF-8 -*-
# @Time    : 2019-06-08 15:51
# @File    : flic.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import scipy.io as io
from antgo.dataflow.dataset import *
import os
import numpy as np

__all__ = ['FLIC']
class FLIC(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(FLIC, self).__init__(train_or_test, dir)
    assert(train_or_test in ['train', 'test'])
    matr = io.loadmat(os.path.join(dir, 'examples.mat'))
    dataset = matr['examples'][0]

    self.ids = []
    self.files = []
    self.coords = []
    self.image_size = []
    self.torso = []

    self.valid_coords = [0,1,2,3,4,5,6,9,12,13,16]
    self.coords_name = ['Nose',
                        'Right Shoulder',
                        'Right Elbow',
                        'Right Wrist',
                        'Right Hip',
                        'Left Shoulder',
                        'Left Elbow',
                        'Left Wrist',
                        'Left Hip',
                        'Left Eye',
                        'Right Eye']
    count = 0
    for data in dataset:
      if train_or_test == 'train' and data[7][0][0]:
        # add train set
        self.ids.append(count)
        self.files.append(data[3][0])
        self.coords.append(data[2]) # 2 x 20
        self.image_size.append(data[4])
        self.torso.append(data[6])

        count += 1
      elif train_or_test == 'test' and data[8][0][0]:
        # add test set
        self.ids.append(count)
        self.files.append(data[3][0])
        self.coords.append(data[2]) # 2 x 20
        self.image_size.append(data[4])
        self.torso.append(data[6])

        count += 1

  @property
  def size(self):
    return len(self.ids)

  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      # idxs = np.arange(len(self.ids))
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)

      for k in idxs:
        img = imread(os.path.join(self.dir, 'images', self.files[k]))
        yield img, {'coords': self.coords[k][:,self.valid_coords],
                    'coords_name': self.coords_name,
                    'id': k,
                    'torso': self.torso[k]}

  def at(self, id):
    img = imread(os.path.join(self.dir, 'images', self.files[id]))
    return img, {'coords': self.coords[id][:,self.valid_coords],
                 'coords_name': self.coords_name,
                'id': id,
                'torso': self.torso[id]}

  def split(self, split_params={}, split_method=''):
    raise NotImplementedError


# encoding=utf-8
# Time: 5/14/17
# File: __init__.py.py
# Author: jian(jian@mltalker.com)
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function



# encoding=utf-8
# @Time    : 17-3-3
# @File    : common.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
from ..utils.serialize import loads,dumps
from ..utils import logger
import uuid
import json
import sys
import tarfile
import tempfile
import re
import requests
from antgo.ant import flags
from antgo.utils.fs import *
from antgo import config
from antgo.ant.utils import *
import yaml
from antgo.utils.utils import *
from datetime import datetime
from antgo.ant.warehouse import *
import antvis.client.mlogger as mlogger
from antgo.context import *
import socket


if sys.version > '3':
  PY3 = True
else:
  PY3 = False

FLAGS = flags.AntFLAGS
Config = config.AntConfig


def _is_open(check_ip, port):
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  try:
    s.connect((check_ip, int(port)))
    s.shutdown(2)
    return True
  except:
    return False


def _pick_idle_port(from_port=40000, check_count=100):
  check_port = from_port
  while check_count:
    if not _is_open('127.0.0.1', check_port):
      break

    logger.warn('port %d is occupied, try to use %d port'%(int(check_port), int(check_port + 1)))

    check_port += 1
    check_count -= 1

    if check_count == 0:
      check_port = None

  if check_port is None:
    logger.warn('couldnt find valid free port')
    exit(-1)

  return check_port

class AntBase(object):
  def __init__(self, ant_name, ant_context=None, ant_token=None, **kwargs):
    self.server_ip = getattr(Config, 'server_ip', 'www.mltalker.com')
    self.http_port = getattr(Config, 'server_port', '8999')
    self.user_token = getattr(Config, 'server_user_token', 'xxx.xxx.xxx.xxx')
    self.http_prefix = 'http'
    self.ant_name = ant_name
    self.app_token = ant_token

    # three key info
    if 'main_file' in kwargs:
      self.main_file = kwargs['main_file']
    if 'main_folder' in kwargs:
      self.main_folder = kwargs['main_folder']
    if 'main_param' in kwargs:
      self.main_param = kwargs['main_param']
    if 'time_stamp' in kwargs:
      self._time_stamp = kwargs['time_stamp']
    else:
      self._time_stamp = timestamp()

    self._proxy = None
    if 'proxy' in kwargs:
      self._proxy = kwargs['proxy']

    self._signature = None
    if 'signature' in kwargs:
      self._signature = kwargs['signature']

    # current pid
    self._pid = str(os.getpid())

    # server flag
    self.app_server = self.__class__.__name__
    if not PY3:
      self.app_server = unicode(self.app_server)

    # parse hardware resource config
    self._running_config = {'GPU_MODEL': '',
                            'GPU_NUM': 0,
                            'GPU_MEM': 0,
                            'CPU_MODEL': '',
                            'CPU_NUM': 0,
                            'CPU_MEM': 0,
                            'OS_PLATFORM': '',
                            'OS_VERSION': '',
                            'SOFTWARE_FRAMEWORK': '',
                            'DATASET': ''}

    self._description_config = {'SHORT_DESCRIPTION': '',
                                'LONG_DESCRIPTION': '',
                                'VERSION': '',
                                'INPUT_NUM': 1,
                                'INPUT_TYPE':[]}

    if ant_context is not None and \
            ant_context.params is not None and \
            ant_context.params._params is not None:
      config_params = ant_context.params._params
      if 'RUNNING_CONFIG' in config_params:
        if 'GPU_MODEL' in config_params['RUNNING_CONFIG']:
          self._running_config['GPU_MODEL'] = config_params['RUNNING_CONFIG']['GPU_MODEL']

        if 'GPU_NUM' in config_params['RUNNING_CONFIG']:
          self._running_config['GPU_NUM'] = config_params['RUNNING_CONFIG']['GPU_NUM']

        if 'GPU_MEM' in config_params['RUNNING_CONFIG']:
          self._running_config['GPU_MEM'] = config_params['RUNNING_CONFIG']['GPU_MEM']

        if 'CPU_MODEL' in config_params['RUNNING_CONFIG']:
          self._running_config['CPU_MODEL'] = config_params['RUNNING_CONFIG']['CPU_MODEL']

        if 'CPU_NUM' in config_params['RUNNING_CONFIG']:
          self._running_config['CPU_NUM'] = config_params['RUNNING_CONFIG']['CPU_NUM']

        if 'CPU_MEM' in config_params['RUNNING_CONFIG']:
          self._running_config['CPU_MEM'] = config_params['RUNNING_CONFIG']['CPU_MEM']

        if 'OS_PLATFORM' in config_params['RUNNING_CONFIG']:
          self._running_config['OS_PLATFORM'] = config_params['RUNNING_CONFIG']['OS_PLATFORM']

        if 'OS_VERSION' in config_params['RUNNING_CONFIG']:
          self._running_config['OS_VERSION'] = config_params['RUNNING_CONFIG']['OS_VERSION']

        if 'SOFTWARE_FRAMEWORK' in config_params['RUNNING_CONFIG']:
          self._running_config['SOFTWARE_FRAMEWORK'] = config_params['RUNNING_CONFIG']['SOFTWARE_FRAMEWORK']

      if 'DESCRIPTION_CONFIG' in config_params:
        if 'SHORT_DESCRIPTION' in config_params['DESCRIPTION_CONFIG']:
          self._description_config['SHORT_DESCRIPTION'] = config_params['DESCRIPTION_CONFIG']['SHORT_DESCRIPTION']

        if 'LONG_DESCRIPTION' in config_params['DESCRIPTION_CONFIG']:
          self._description_config['LONG_DESCRIPTION'] = config_params['DESCRIPTION_CONFIG']['LONG_DESCRIPTION']

        if 'VERSION' in config_params['DESCRIPTION_CONFIG']:
          self._description_config['VERSION'] = config_params['DESCRIPTION_CONFIG']['VERSION']

        if 'INPUT_NUM' in config_params['DESCRIPTION_CONFIG']:
          self._description_config['INPUT_NUM'] = config_params['DESCRIPTION_CONFIG']['INPUT_NUM']

        if 'INPUT_TYPE' in config_params['DESCRIPTION_CONFIG']:
          self._description_config['INPUT_TYPE'] = config_params['DESCRIPTION_CONFIG']['INPUT_TYPE']

    self._running_platform = kwargs.get('running_platform', 'local')    # local, cloud

    # global context
    self.ant_context = None
    if ant_context is not None:
      self.ant_context = ant_context
      self.ant_context.ant = self

    self.experiment_uuid = \
      '%s-%s-%s' % (str(uuid.uuid4()),
                    str(uuid.uuid4()),
                    datetime.fromtimestamp(timestamp()).strftime('%Y%m%d-%H%M%S-%f'))
    self.context.experiment_uuid = self.experiment_uuid

    if self.app_server in ["AntChallenge","AntTrain"]:
      if self.app_token is not None:
        # 任务模式，在dashboard上创建实验记录
        mlogger.config(ip=self.server_ip,
                      port=(int)(self.http_port),
                      project=self.ant_name,
                      experiment=self.ant_name,
                      token=self.app_token)

        if mlogger.getEnv().dashboard.experiment_uuid is not None:
          self.experiment_uuid = mlogger.getEnv().dashboard.experiment_uuid
          self.context.experiment_uuid = mlogger.getEnv().dashboard.experiment_uuid
      else:
        # 非任务模式，基于user token与dashboard进行通信
        mlogger.config(ip=self.server_ip,
                      port=(int)(self.http_port),
                      project=self.ant_name,
                      experiment='exp',
                      token=self.user_token)
        if mlogger.getEnv().dashboard.experiment_uuid is not None:
          self.experiment_uuid = mlogger.getEnv().dashboard.experiment_uuid
          self.context.experiment_uuid = mlogger.getEnv().dashboard.experiment_uuid

  @property
  def pid(self):
    return self._pid

  @pid.setter
  def pid(self, val):
    self._pid = val

  @property
  def running_config(self):
    return self._running_config

  @property
  def description_config(self):
    return self._description_config

  @property
  def running_platform(self):
    return self._running_platform

  def package_codebase(self, prefix='qiniu', target_path='', signature='123'):
    logger.info('package code envoriment')
    if self.app_token is None:
      if not os.path.exists(os.path.join(self.main_folder, FLAGS.task())):
        shutil.copy(os.path.join(Config.task_factory, FLAGS.task()), os.path.join(self.main_folder))

    tar_shell = 'tar -czf - * | openssl enc -e -aes256 -out %s.tar.gz -k %s' % (self.name, signature)
    subprocess.call(tar_shell, shell=True, cwd=self.main_folder)

    # logger.info('finish package')
    # if prefix == 'qiniu':
    #   logger.info('upload codebase package')
    #   qiniu_address = qiniu_upload(os.path.join(self.main_folder, '%s.tar.gz'%self.name),
    #                                bucket='experiment',
    #                                max_size=100)
    #   # clear
    #   os.remove(os.path.join(self.main_folder, '%s.tar.gz' % self.name))
    #   return qiniu_address
    # elif prefix == 'ipfs':
    #   pass
    # elif prefix == 'baidu':
    #   pass
    # elif prefix.startswith('ssh') or prefix.startswith('scp'):
    #   nodes = prefix.replace('scp:', '')
    #   node_ip_list = nodes.split(',')
    #   for ip in node_ip_list:
    #     if ip=='127.0.0.1' or ip=='localhost':
    #       continue
    #
    #     logger.info('deploy code at %s'%ip)
    #     try:
    #       cmd_str = 'ssh %s %s' % (ip, 'mkdir -p %s'%target_path)
    #       logger.info('execute %s' % cmd_str)
    #       subprocess.call(cmd_str, shell=True)
    #     except:
    #       pass
    #
    #     try:
    #       cmd_str = 'scp %s %s:%s' % (os.path.join(self.main_folder, '%s.tar.gz' % self.name), ip, target_path)
    #       logger.info('execute %s' % cmd_str)
    #       subprocess.call(cmd_str, shell=True)
    #     except:
    #       logger.error('couldnt distribute code base to %s' % ip)
    #       exit(-1)
    #
    #   # clear
    #   os.remove(os.path.join(self.main_folder, '%s.tar.gz' % self.name))

    return '%s.tar.gz' % self.name

  def download(self, source_path, target_path=None, target_name=None, archive=None):
    if target_path is None:
      target_path = os.curdir

    is_that = re.match('^((https|http|ftp|rtsp|mms)?://)', source_path)
    if is_that is not None:
      download(source_path, target_path, fname=target_name)

      is_gz = re.match('.*\.gz', target_name)
      if is_gz is not None:
        if archive is not None:
          extracted_path = os.path.join(target_path, archive)
        else:
          extracted_path = target_path

        if not os.path.exists(extracted_path):
          os.makedirs(extracted_path)

        tar = tarfile.open(os.path.join(target_path, target_name))
        tar.extractall(extracted_path)
        tar.close()
        target_path = extracted_path

    return target_path

  @property
  def stage(self):
    return self.context.stage

  @stage.setter
  def stage(self, val):
    self.context.stage = val

  @property
  def token(self):
    return self.app_token
  @token.setter
  def token(self, val):
    self.app_token = val

  @property
  def name(self):
    return self.ant_name

  @property
  def context(self):
    return self.ant_context

  @context.setter
  def context(self, val):
    self.ant_context = val
    self.ant_context.ant = self

  @property
  def proxy(self):
    return self._proxy

  @property
  def signature(self):
    return self._signature

  @property
  def time_stamp(self):
    return self._time_stamp
  
  def reset(self):
    if self.pid != str(os.getpid()):
      # 1.step reset process pid
      self.pid = str(os.getpid())

      # 2.step update context
      ctx = main_context(self.main_file, self.main_folder)
      if self.main_param is not None:
        main_config_path = os.path.join(self.main_folder, self.main_param)
        params = yaml.load(open(main_config_path, 'r'))
        ctx.params = params
      
      if self.context.from_experiment is not None:
        ctx.from_experiment = self.context.from_experiment
      
      self.context = ctx

      # 3.step update dashboard configure
      # self.context.dashboard.configure(dashboard_ip=self.server_ip,
      #                                  dashboard_port=int(self.http_port),
      #                                  token=self.app_token,
      #                                  experiment_uuid=self.experiment_uuid,
      #                                  experiment_name=self.ant_name,
      #                                  server=self.app_server)
      # mlogger.config(ip=self.server_ip,
      #                port=int(self.http_port),
      #                token=self.app_token,
      #                )
      # mlogger.activate(project='', experiment='')